// Template for an interpose library that can be used to hook C/C++ methods
// and/or swizzle Objective-C methods for debugging/reverse-engineering.
//
// A number of methods are provided to be called from your hooks that make use
// of Apple's Symbolication framework (which though undocumented is heavily
// used by Apple utilities such as crashreporterd and dtrace).  Particularly
// useful is PrintStackTrace().
//
// Once the interpose library is built, use it as follows:
//
// A) From a Terminal prompt:
//    1) export DYLD_INSERT_LIBRARIES=/full/path/to/interpose.dylib
//    2) /path/to/application
//
// B) From gdb:
//    1) set DYLD_INSERT_LIBRARIES /full/path/to/interpose.dylib
//    2) run

#include <dlfcn.h>
#include <pthread.h>
#include <libproc.h>
#include <stdarg.h>
#include <time.h>
#import <Cocoa/Cocoa.h>
#import <Carbon/Carbon.h>
#import <objc/Object.h>

class nsAutoreleasePool {
public:
    nsAutoreleasePool()
    {
        mLocalPool = [[NSAutoreleasePool alloc] init];
    }
    ~nsAutoreleasePool()
    {
        [mLocalPool release];
    }
private:
    NSAutoreleasePool *mLocalPool;
};

char *GetOwnerName(void *address);
char *GetAddressName(void *address);
void PrintAddress(void *address);
void PrintStackTrace(NSArray *addresses = nil);
BOOL SwizzleMethods(Class aClass, SEL orgMethod, SEL posedMethod, BOOL classMethods);

char gProcPath[PROC_PIDPATHINFO_MAXSIZE] = {0};

static void MaybeGetProcPath()
{
  if (gProcPath[0]) {
    return;
  }
  proc_pidpath(getpid(), gProcPath, sizeof(gProcPath) - 1);
}

static void LogWithFormatV(bool decorate, CFStringRef format, va_list args)
{
  MaybeGetProcPath();

  CFStringRef message = CFStringCreateWithFormatAndArguments(kCFAllocatorDefault, NULL,
                                                             format, args);

  int msgLength = CFStringGetMaximumSizeForEncoding(CFStringGetLength(message),
                                                    kCFStringEncodingUTF8);
  char *msgUTF8 = (char *) calloc(msgLength + 1, 1);
  CFStringGetCString(message, msgUTF8, msgLength, kCFStringEncodingUTF8);
  CFRelease(message);

  char *finished = (char *) calloc(msgLength + 1024, 1);
  const time_t currentTime = time(NULL);
  char timestamp[30] = {0};
  ctime_r(&currentTime, timestamp);
  timestamp[strlen(timestamp) - 1] = 0;
  if (decorate) {
    sprintf(finished, "(%s) %s[%u] %s\n",
            timestamp, gProcPath, getpid(), msgUTF8);
  } else {
    sprintf(finished, "%s\n", msgUTF8);
  }
  free(msgUTF8);

  fputs(finished, stdout);

  free(finished);
}

static void LogWithFormat(bool decorate, const char *format, ...)
{
  va_list args;
  va_start(args, format);
  CFStringRef formatCFSTR = CFStringCreateWithCString(kCFAllocatorDefault, format,
                                                      kCFStringEncodingUTF8);
  LogWithFormatV(decorate, formatCFSTR, args);
  CFRelease(formatCFSTR);
  va_end(args);
}

extern "C" void interposelib_LogWithFormatV(bool decorate, const char *format, va_list args)
{
  CFStringRef formatCFSTR = CFStringCreateWithCString(kCFAllocatorDefault, format,
                                                      kCFStringEncodingUTF8);
  LogWithFormatV(decorate, formatCFSTR, args);
  CFRelease(formatCFSTR);
}

extern "C" void interposelib_PrintStackTrace()
{
  PrintStackTrace();
}

Class gSymbolicatorClass = nil;
static BOOL gSymbolicationLoaded = NO;
static BOOL loadSymbolication()
{
  if (gSymbolicationLoaded) {
    return YES;
  }
  // Needed for printing stack traces.  Available on OS X 10.5 and up.
  if (dlopen("/System/Library/PrivateFrameworks/Symbolication.framework/Symbolication", RTLD_LAZY)) {
    gSymbolicationLoaded = YES;
    Class VMUDebugMapExtractorClass = ::NSClassFromString(@"VMUDebugMapExtractor");
    SwizzleMethods(VMUDebugMapExtractorClass, @selector(debugMapExtractorWithMachOHeader:),
                   @selector(VMUDebugMapExtractor_debugMapExtractorWithMachOHeader:), YES);
    gSymbolicatorClass = ::NSClassFromString(@"VMUSymbolicator");

    // Swizzle other methods here
  }

  return gSymbolicationLoaded;
}

pthread_t gMainThreadID = 0;

bool IsMainThread()
{
  return (gMainThreadID == pthread_self());
}

static int Hooked_pthread_once(pthread_once_t *once_control, void (*init_routine)(void))
{
  int retval = pthread_once(once_control, init_routine);
  if (!gMainThreadID) {
    gMainThreadID = pthread_self();
  }
  return retval;
}

extern "C" void *NSPushAutoreleasePool();

static void *Hooked_NSPushAutoreleasePool()
{
  void *retval = NSPushAutoreleasePool();
  if (IsMainThread()) {
    loadSymbolication();
  }
  return retval;
}

// Put other hooked methods and swizzled classes here

extern "C" int32_t decode_data(void *arg1, uint32_t arg2, uint32_t arg3, void *arg4, uint32_t arg5, void *arg6,
                               uint32_t arg7, void *arg8, uint32_t arg9);

static int32_t Hooked_decode_data(void *arg1, uint32_t arg2, uint32_t arg3, void *arg4, uint32_t arg5, void *arg6,
                                  uint32_t arg7, void *arg8, uint32_t arg9)
{
  LogWithFormat(true, "decode_data(1): arg1 %p, arg2 %u, arg3 %u, arg4 %p, arg5 %u, arg6 %p, arg7 %u, arg8 %p, arg9 %u",
                arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9);
  int32_t retval = decode_data(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9);
  LogWithFormat(true, "decode_data(2): arg1 %p, arg2 %u, arg3 %u, arg4 %p, arg5 %u, arg6 %p, arg7 %u, arg8 %p, arg9 %u, returning %i",
                arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, retval);
  return retval;
}

static CGImageRef Hooked_CGImageCreate(size_t width, size_t height,
                                       size_t bitsPerComponent, size_t bitsPerPixel,
                                       size_t bytesPerRow,
                                       CGColorSpaceRef colorspace,
                                       CGBitmapInfo bitmapInfo,
                                       CGDataProviderRef provider,
                                       const CGFloat decode[],
                                       bool shouldInterpolate,
                                       CGColorRenderingIntent intent)
{
  LogWithFormat(true, "CGImageCreate(1): width %u, height %u, bitsPerComponent %u, bitsPerPixel %u, bytesPerRow %u, colorspace %p, bitmapInfo %p, provider %p, decode %f, shouldInterpolate %i, intent %i",
                width, height, bitsPerComponent, bitsPerPixel, bytesPerRow,
                colorspace, (void *) bitmapInfo, provider, decode ? decode[0] : 0,
                shouldInterpolate, intent);
  CGImageRef retval = CGImageCreate(width, height,
                                    bitsPerComponent, bitsPerPixel, bytesPerRow,
                                    colorspace, bitmapInfo, provider, decode,
                                    shouldInterpolate, intent);
  LogWithFormat(true, "CGImageCreate(2): width %u, height %u, bitsPerComponent %u, bitsPerPixel %u, bytesPerRow %u, colorspace %p, bitmapInfo %p, provider %p, decode %f, shouldInterpolate %i, intent %i, returning %p",
                width, height, bitsPerComponent, bitsPerPixel, bytesPerRow,
                colorspace, (void *) bitmapInfo, provider, decode ? decode[0] : 0,
                shouldInterpolate, intent, retval);
  PrintStackTrace();
  return retval;
}

extern "C" void*_CFRuntimeCreateInstance(void *allocator, CFTypeID type, CFIndex size, int32_t arg4);

static void *Hooked__CFRuntimeCreateInstance(void *allocator, CFTypeID type, CFIndex size, int32_t arg4)
{
  bool log = (type == CGImageGetTypeID());
  if (log) {
    LogWithFormat(true, "_CFRuntimeCreateInstance(1): allocator %p, type %lu, size %li, arg4 %i",
                  allocator, type, size, arg4);
  }
  void *retval = _CFRuntimeCreateInstance(allocator, type, size, arg4);
  if (log) {
    LogWithFormat(true, "_CFRuntimeCreateInstance(2): allocator %p, type %lu, size %li, arg4 %i, returning %p",
                  allocator, type, size, arg4, retval);
    PrintStackTrace();
  }
  return retval;
}

extern "C" void CGImageSetProperty(CGImageRef image, CFStringRef property, CFTypeRef value);
extern "C" CFTypeRef CGImageGetProperty(CGImageRef image, CFStringRef property);

static CFTypeRef Hooked_CFRetain(CFTypeRef obj)
{
  if (obj && (CFGetTypeID(obj) == CGImageGetTypeID()) &&
      CGImageGetProperty((CGImageRef)obj, CFSTR("royal.pita.com")))
  {
    LogWithFormat(true, "CFRetain(): image %p", obj);
    PrintStackTrace();
  }
  CFTypeRef retval = CFRetain(obj);
  return retval;
}

static void Hooked_CFRelease(CFTypeRef obj)
{
  if (obj && (CFGetTypeID(obj) == CGImageGetTypeID()) &&
      CGImageGetProperty((CGImageRef)obj, CFSTR("royal.pita.com")))
  {
    LogWithFormat(true, "CFRelease(): image %p", obj);
    PrintStackTrace();
  }
  CFRelease(obj);
}

extern "C" uint32_t CGSImageStreamRead(void *arg1, long arg2, long arg3);

static uint32_t Hooked_CGSImageStreamRead(void *arg1, long arg2, long arg3)
{
  LogWithFormat(true, "CGSImageStreamRead(1): arg1 %p, arg2 %li, arg3 %li",
                arg1, arg2, arg3);
  uint32_t retval = CGSImageStreamRead(arg1, arg2, arg3);
  LogWithFormat(true, "CGSImageStreamRead(2): arg1 %p, arg2 %li, arg3 %li, returning %i",
                arg1, arg2, arg3, retval);
  return retval;
}

extern "C" void *CGSImageStreamLock(CGImageRef image, uint32_t arg2, uint32_t arg3, uint32_t arg4, uint32 arg5, uint32 arg6,
                                    long arg7, uint32 arg8);

static void *Hooked_CGSImageStreamLock(CGImageRef image, uint32_t arg2, uint32_t arg3, uint32_t arg4, uint32 arg5, uint32 arg6,
                                       long arg7, uint32 arg8)
{
  LogWithFormat(true, "CGSImageStreamLock(1): image %p (width %u, height %u, retainCount %li), arg2 %u, arg3 %u, arg4 %u, arg5 %u, arg6 %u, arg7 %li, arg8 %u",
                image, CGImageGetWidth(image), CGImageGetHeight(image), CFGetRetainCount(image),
                arg2, arg3, arg4, arg5, arg6, arg7, arg8);
  void *retval = CGSImageStreamLock(image, arg2, arg3, arg4, arg5, arg6, arg7, arg8);
  LogWithFormat(true, "CGSImageStreamLock(2): image %p (width %u, height %u, retainCount %li), arg2 %u, arg3 %u, arg4 %u, arg5 %u, arg6 %u, arg7 %li, arg8 %u, returning %p",
                image, CGImageGetWidth(image), CGImageGetHeight(image), CFGetRetainCount(image),
                arg2, arg3, arg4, arg5, arg6, arg7, arg8, retval);
  return retval;
}

extern "C" void CGSImageStreamUnlock(void *arg1);

static void Hooked_CGSImageStreamUnlock(void *arg1)
{
  LogWithFormat(true, "CGSImageStreamUnlock(): arg1 %p", arg1);
  CGSImageStreamUnlock(arg1);
}

#pragma mark -

struct interpose_substitution {
  const void* replacement;
  const void* original;
};

#define INTERPOSE_FUNCTION(function) \
    { reinterpret_cast<const void*>(Hooked_##function), \
      reinterpret_cast<const void*>(function) }

__attribute__((used)) static const interpose_substitution substitutions[]
    __attribute__((section("__DATA, __interpose"))) = {
  INTERPOSE_FUNCTION(pthread_once),
  INTERPOSE_FUNCTION(NSPushAutoreleasePool),
  //INTERPOSE_FUNCTION(decode_data),
  //INTERPOSE_FUNCTION(CGImageCreate),
  //INTERPOSE_FUNCTION(_CFRuntimeCreateInstance),
  INTERPOSE_FUNCTION(CFRetain),
  INTERPOSE_FUNCTION(CFRelease),
  INTERPOSE_FUNCTION(CGSImageStreamRead),
  INTERPOSE_FUNCTION(CGSImageStreamLock),
  INTERPOSE_FUNCTION(CGSImageStreamUnlock),
};

// What follows are definitions of the Symbolication interfaces we use to get
// stack traces.  The Symbolication framework is an undocumented, private
// framework available on OS X 10.5 and up.  It's used by Apple utilities
// such as dtrace.  http://www.cocoadev.com/index.pl?StackTraces has basic
// information on how to use the Symbolication framework.

@interface VMUAddressRange : NSObject <NSCoding>
{
}

@end

@interface VMUSymbol : VMUAddressRange <NSCopying>
{
}

- (NSString *)name;
- (NSString *)description;
- (id)text;

@end

@interface VMUSymbolOwner : NSObject <NSCopying>
{
}

- (NSString *)name;

@end

// We must override the debugMapExtractorWithMachOHeader method:  Otherwise
// the Symbolication framework can take a minute or more to initialize!!
// We don't need the information in the "debug maps" (Dwarf or dsym debugging
// information) to get simple stack traces.
@interface NSObject (VMUDebugMapExtractorMethodSwizzling)
+ (id)VMUDebugMapExtractor_debugMapExtractorWithMachOHeader:(id)machHeader;
@end

@implementation NSObject (VMUDebugMapExtractorMethodSwizzling)

+ (id)VMUDebugMapExtractor_debugMapExtractorWithMachOHeader:(id)machHeader
{
  return nil;
}

@end

@interface VMUSymbolicator : NSObject
{
}

+ (VMUSymbolicator *)symbolicatorForPid:(int)pid;
- (VMUSymbol *)symbolForAddress:(unsigned long long)address;
- (VMUSymbolOwner *)symbolOwnerForAddress:(unsigned long long)address;

@end

@interface VMUSymbolicator (PerformanceOptimization)
- (id)symbolOwnerNameForAddress:(unsigned long long)fp8;
@end

char *GetOwnerName(void *address)
{
  static char holder[1024] = {0};
  if (!loadSymbolication()) {
    holder[0] = 0;
    return holder;
  }
  VMUSymbolicator *symbolicator = [gSymbolicatorClass symbolicatorForPid:getpid()];
  if (!symbolicator) {
    holder[0] = 0;
    return holder;
  }

  VMUSymbol *symbol =
    [symbolicator symbolForAddress:(unsigned long long)address];
  NSString *ownerName = nil;
  if ([symbolicator respondsToSelector:@selector(symbolOwnerNameForAddress:)]) {
    ownerName =
      [symbolicator symbolOwnerNameForAddress:(unsigned long long)address];
  } else {
    VMUSymbolOwner *owner =
      [symbolicator symbolOwnerForAddress:(unsigned long long)address];
    ownerName = [owner name];
  }
  if (ownerName) {
    strcpy(holder, [ownerName UTF8String]);
    return holder;
  }
  holder[0] = 0;
  return holder;
}

char *GetAddressName(void *address)
{
  static char holder[1024] = {0};
  if (!loadSymbolication()) {
    holder[0] = 0;
    return holder;
  }
  VMUSymbolicator *symbolicator = [gSymbolicatorClass symbolicatorForPid:getpid()];
  if (!symbolicator) {
    holder[0] = 0;
    return holder;
  }

  VMUSymbol *symbol =
    [symbolicator symbolForAddress:(unsigned long long)address];
  if (symbol && [symbol name]) {
    strcpy(holder, [[symbol name] UTF8String]);
    return holder;
  }
  holder[0] = 0;
  return holder;
}

void PrintAddress(void *address)
{
  if (!loadSymbolication()) {
    return;
  }
  VMUSymbolicator *symbolicator = [gSymbolicatorClass symbolicatorForPid:getpid()];
  if (!symbolicator) {
    return;
  }

  VMUSymbol *symbol =
    [symbolicator symbolForAddress:(unsigned long long)address];
  NSString *ownerName = nil;
  if ([symbolicator respondsToSelector:@selector(symbolOwnerNameForAddress:)]) {
    ownerName =
      [symbolicator symbolOwnerNameForAddress:(unsigned long long)address];
  } else {
    VMUSymbolOwner *owner =
      [symbolicator symbolOwnerForAddress:(unsigned long long)address];
    ownerName = [owner name];
  }
  LogWithFormat(false, "    (%@) %@, %p", ownerName, [symbol name], address);
}

void PrintStackTrace(NSArray *addresses)
{
  if (!loadSymbolication()) {
    return;
  }
  VMUSymbolicator *symbolicator = [gSymbolicatorClass symbolicatorForPid:getpid()];
  if (!symbolicator) {
    return;
  }

  if (!addresses) {
    addresses = [NSThread callStackReturnAddresses];
  }
  NSUInteger count = [addresses count];
  for (NSUInteger i = 0; i < count; ++i) {
    NSNumber *item = (NSNumber *) [addresses objectAtIndex:i];
    VMUSymbol *symbol =
      [symbolicator symbolForAddress:[item unsignedLongLongValue]];
    NSString *ownerName = nil;
    if ([symbolicator respondsToSelector:@selector(symbolOwnerNameForAddress:)]) {
      ownerName =
        [symbolicator symbolOwnerNameForAddress:[item unsignedLongLongValue]];
    } else {
      VMUSymbolOwner *owner =
        [symbolicator symbolOwnerForAddress:[item unsignedLongLongValue]];
      ownerName = [owner name];
    }
    LogWithFormat(false, "    (%@) %@, %p",
                  ownerName, [symbol name], (void *)[item unsignedLongLongValue]);
  }
}

BOOL SwizzleMethods(Class aClass, SEL orgMethod, SEL posedMethod, BOOL classMethods)
{
  Method original = nil;
  Method posed = nil;

  if (classMethods) {
    original = class_getClassMethod(aClass, orgMethod);
    posed = class_getClassMethod(aClass, posedMethod);
  } else {
    original = class_getInstanceMethod(aClass, orgMethod);
    posed = class_getInstanceMethod(aClass, posedMethod);
  }

  if (!original || !posed)
    return NO;

  method_exchangeImplementations(original, posed);

  return YES;
}
